"""BHFIELD sphere wrapper"""
import os
import pathlib
import subprocess as sp
import tempfile
import warnings

import numpy as np

from .fetch import get_binary


class BHFIELDExecutionError(BaseException):
    """Raised when BHFIELD fails"""
    pass


def clear_temp(wdir, rmdir=True):
    """Remove all files in wdir"""
    wdir = pathlib.Path(wdir)
    extensions = ["*.log", "*.dat", "*.txt"]

    for ext in extensions:
        for ff in wdir.glob(ext):
            ff.unlink()

    if rmdir:
        wdir.rmdir()


def load_field(wdir, shape_grid):
    """Extract electric field from simulation file "V_0Ereim.dat"

    Parameters
    ----------
    wdir: str or pathlib.Path
        path to the working directory
    shape_grid: tuple of ints
        the shape of the simulation data grid

    Notes
    -----
    These are the files present in the working directory

    bhfield.log:
        log file containing scattering coeffs etc
        (as well as definitions of the output fields)

    bhdebug.log:
        extra information (checking roundoff errors etc)

    E_0allf.dat:
        E2 = EFSQ[dimensionless] = (relative) squared amplitude
        of complex E-field; Ec*(Ec^*) / E0**2

    E_1core.dat:
        E2 inside the core

    E_2coat.dat:
        E2 inside the coating

    E_3exte.dat:
        E2 at the outside of the sphere

    U_*.dat:
        U = UABS[F m-1 s-1] is the (relative) absorbed energy per
        unit volume and time; Ua [W m-3] / E0**2

    EU_zax.txt:
        E2(0,0,z) and U(0,0,z) along z-axis; it may be blank if
        the grid does not include such points

    V_0Eelli.dat:
        vector electric field; vibration ellipse (major & minor axes),
        ellipticity, azimuth[deg], p-a angle (phi)[deg], handedness
        angle[deg] & handedness

    V_0Ereim.dat: vector electric field;
        snapshots [Re (t=0), Im (t=period/4)]

    V_0Helli.dat:
        vector magnetic field; vibration ellipse (major & minor axes),
        ellipticity, azimuth[deg], p-a angle (phi)[deg],
        E-&H-phase dif[deg], handedness angle[deg] & handedness

    V_0Hreim.dat: vector magnetic field;
        snapshots [Re (t=0), Im (t=period/4)]

    V_0Poynt.dat:
        Poynting vector <S>, EH angle, optical irradiance (intensity)
        (norm<S>), I(plane), -div<S> (1st-3rd), UABS & DIVSR

    V_1*.dat:
        vector fields inside the core

    V_2*.dat:
        vector fields inside the coating

    V_3*.dat:
        vector fields at the outside of the sphere
    """
    wdir = pathlib.Path(wdir)
    check_simulation(wdir)
    field_file = wdir / "V_0Ereim.dat"

    a = np.loadtxt(field_file)

    assert shape_grid[0] == int(
        shape_grid[0]), "resulting x-size is not an integer"
    assert shape_grid[1] == int(
        shape_grid[1]), "resulting y-size is not an integer"

    Ex = a[:, 3] + 1j * a[:, 6]
    # Ey = a[:,4] + 1j*a[:,7]
    # Ez = a[:,5] + 1j*a[:,8]

    Exend = Ex.reshape((shape_grid[1], shape_grid[0])).transpose()
    return Exend


def simulate_sphere(radius_sphere_um=2.5,
                    size_simulation_um=[7, 7],
                    shape_grid=(50, 50),
                    refractive_index_medium=1.0,
                    refractive_index_sphere=1.01,
                    measurement_position_um=2.5,
                    wavelength_nm=500,
                    offset_x_um=0,
                    offset_y_um=0,
                    arp=True):
    """
    Parameters
    ----------
    radius_sphere_um : float
        radius of sphere in um
    size_simulation_um : list of floats
        Size of simulation volume in lateral dimension in um.
        If a float is given, then a square simulation size is assumed. If
        a tuple is given, then a rectangular shape is assumed.
    shape_grid : tuple of ints
        grid points in each lateral dimension.
        If a float is given, then a square simulation size is assumed. If
        a tuple is given, then a rectangular shape is assumed.
    refractive_index : float
        RI of sphere
    measurement_position_um : float
        axial measurement position in um
    wavelength_nm : float
        light wavelength in nanometers
    offset_x_um : float
        x coordinate of center of the sphere in um
    offset_y_um : float
        y coordinate of center of the sphere in um
    arp: bool
        Use arbitrary precision
    """
    wavelength_um = wavelength_nm / 1000

    # size simulation tuple
    sizeum = list(size_simulation_um)
    shape_grid = list(shape_grid)

    # The size of the simulation must be zero
    # if there is only one grid point.
    if shape_grid[0] == 1:
        sizeum[0] = 0
    if shape_grid[1] == 1:
        sizeum[1] = 0

    assert np.allclose(shape_grid[0], int(
        shape_grid[0])), "resulting x-size is not an integer"
    assert np.allclose(shape_grid[1], int(
        shape_grid[1])), "resulting y-size is not an integer"

    while True:
        # create temp dir
        wdir = tempfile.mkdtemp(prefix="qpsphere_bhfield_")

        try:
            run_simulation(wdir=wdir,
                           arp=arp,
                           wl=wavelength_um,
                           r_core=radius_sphere_um,
                           r_coat=radius_sphere_um,
                           n_grid_x=int(shape_grid[0]),
                           xspan_min=-sizeum[0] / 2 - offset_x_um,
                           xspan_max=sizeum[0] / 2 - offset_x_um,
                           n_grid_y=int(shape_grid[1]),
                           yspan_min=-sizeum[1] / 2 - offset_y_um,
                           yspan_max=sizeum[1] / 2 - offset_y_um,
                           n_grid_z=1,
                           zspan_min=measurement_position_um,
                           zspan_max=measurement_position_um,
                           case="other",
                           Kreibig=0,
                           n_med=refractive_index_medium,
                           n_core=refractive_index_sphere,
                           k_core=0,
                           n_coat=refractive_index_medium,
                           k_coat=0)
        except BaseException:
            if arp:
                raise
            else:
                msg = "bhfield: Standard precision failed. " \
                      + "Retrying with arbitrary precision."
                warnings.warn(msg)
                arp = True
            clear_temp(wdir=wdir)
        else:
            break

    result = load_field(wdir=wdir, shape_grid=shape_grid)
    clear_temp(wdir=wdir)

    return result


def run_simulation(wdir, arp=True, **kwargs):
    """
    Example
    -------
    100-nm silica sphere with 10-nm thick Ag coating,
    embedded in water; arprec 20 digits; illuminated with YAG (1064nm);
    scan xz plane (21x21, +-200nm)

    bhfield-arp-db.exe mpdigit wl r_core r_coat
                       n_grid_x xspan_min xspan_max
                       n_grid_y yspan_min yspan_max
                       n_grid_z zspan_min zspan_max
                       case  Kreibig
                       [n_med n_core k_core n_coat k_coat (case=other)]

    bhfield-arp-db.exe 20 1.064 0.050 0.060
                       21 -0.2 0.2
                       1 0 0
                       21 -0.2 0.2
                       other 0
                       1.3205 1.53413 0 0.565838 7.23262


    Explanation of parameters
    -------------------------
    mpdigit:
        arprec's number of precision digits;
        increase it to overcome round-off errors
    wl[um]:
        light wavelength in vacuum
    r_core[um], r_coat[um]:
        core & coat radii
    n_grid_x xspan_min[um] xspan_max[um]:
        number & span of grid points for field computation; x span
    n_grid_y yspan_min[um] yspan_max[um]:
        y span
    n_grid_z zspan_min[um] zspan_max[um]:
        z span
    Kreibig:
        Kreibig mean free path correction for Ag (0.0 - 1.0)
    case:
        nanoshell/liposome/HPC/barber/other
    n_med n_core k_core n_coat k_coat (case=other only):
        refractive indices of medium (real), core & coat (n, k)


    If `case=other`, complex refractive indices
    (n, k at the particular wavelength) must be specified.
    Otherwise (case = nanoshell etc) the medium/core/coat
    materials are predefined and the n,k values
    are taken from the data file (Ag_palik.nk etc).
    The latter reflects our own interest and is intended
    for use in our lab, so general users may not find it useful :-)
    """
    wdir = pathlib.Path(wdir)
    cmd = "{pathbhfield} {mpdigit} {wl:f} {r_core:f} {r_coat:f} " \
          + "{n_grid_x:d} {xspan_min:f} {xspan_max:f} " \
          + "{n_grid_y:d} {yspan_min:f} {yspan_max:f} " \
          + "{n_grid_z:d} {zspan_min:f} {zspan_max:f} " \
          + "{case} {Kreibig:f} {n_med:f} {n_core:f} {k_core:f} " \
          + "{n_coat:f} {k_coat:f}"
    old_dir = pathlib.Path.cwd()
    os.chdir(wdir)

    kwargs["pathbhfield"] = get_binary(arp=arp)

    if arp:
        kwargs["mpdigit"] = 16
    else:
        kwargs["mpdigit"] = ""

    # run simulation with kwargs
    sp.check_output(cmd.format(**kwargs), shell=True)

    # Go back to orgignal directory before checking (checking might fail)
    os.chdir(old_dir)

    # Check bhdebug.txt to make sure that you specify enough digits to
    # overcome roundoff errors.
    check_simulation(wdir)


def check_simulation(wdir):
    """
    Check bhdebug.txt to make sure that you specify enough digits to
    overcome roundoff errors.
    """
    wdir = pathlib.Path(wdir)
    field = wdir / "V_0Ereim.dat"
    if not (field.exists() and
            field.stat().st_size > 130):
        msg = "Output {} does not exist or is too small!".format(field)
        raise BHFIELDExecutionError(msg)



import numpy as np

from qpsphere.models import mie


def test_basic():
    data = np.array([0.61718511, 0.63344636, 0.39500855,
                     0.62834482, 0.5612372, 0.24405929,
                     0.39808633, 0.24597916, -0.00504312])

    qpi = mie(grid_size=(3, 3),
              center=(0, 0),
              pixel_size=2e-6)
    assert np.allclose(data, qpi.pha.flatten())


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import numbers
import numpy as np

from .mod_mie import mie
from .mod_mie_avg import mie_avg
from .mod_proj import projection
from .mod_rytov import rytov
from .mod_rytov_sc import rytov_sc

model_dict = {"mie": mie,
              "mie-avg": mie_avg,
              "projection": projection,
              "rytov": rytov,
              "rytov-sc": rytov_sc,
              }

#: available light-scattering models
available = sorted(list(model_dict.keys()))


def simulate(radius=5e-6, sphere_index=1.339, medium_index=1.333,
             wavelength=550e-9, grid_size=(80, 80), model="projection",
             pixel_size=None, center=None):
    """Simulate scattering at a sphere

    Parameters
    ----------
    radius: float
        Radius of the sphere [m]
    sphere_index: float
        Refractive index of the object
    medium_index: float
        Refractive index of the surrounding medium
    wavelength: float
        Vacuum wavelength of the imaging light [m]
    grid_size: tuple of int
        Resulting image size in x and y [px]
    model: str
        Sphere model to use (see :const:`available`)
    pixel_size: float or None
        Pixel size [m]; if set to `None` the pixel size is
        chosen such that the radius fits at least three to
        four times into the grid.
    center: tuple of floats or None
        Center position in image coordinates [px]; if set to
        None, the center of the image (grid_size - 1)/2 is
        used.

    Returns
    -------
    qpi: qpimage.QPImage
        Quantitative phase data set
    """
    if isinstance(grid_size, numbers.Integral):
        # default to square-shape grid
        grid_size = (grid_size, grid_size)
    if pixel_size is None:
        # select simulation automatically
        rl = radius / wavelength
        if rl < 5:
            # a lot of diffraction artifacts may occur;
            # use 4x radius to capture the full field
            fact = 4
        elif rl >= 5 and rl <= 10:
            # linearly decrease towards 3x radius
            fact = 4 - (rl - 5) / 5
        else:
            # not many diffraction artifacts occur;
            # 3x radius is enough and allows to
            # simulate larger radii with BHFIELD
            fact = 3
        pixel_size = fact * radius / np.min(grid_size)

    if center is None:
        center = (np.array(grid_size) - 1) / 2

    model = model_dict[model]
    qpi = model(radius=radius,
                sphere_index=sphere_index,
                medium_index=medium_index,
                wavelength=wavelength,
                pixel_size=pixel_size,
                grid_size=grid_size,
                center=center)

    return qpi



from ._version import version as __version__  # noqa: F401
from . import edgefit
from . import imagefit
from . import models  # noqa: F401
from .models import simulate  # noqa: F401
from . import util  # noqa: F401


def analyze(qpi, r0, method="edge", model="projection", edgekw={}, imagekw={},
            ret_center=False, ret_pha_offset=False, ret_qpi=False):
    """Determine refractive index and radius of a spherical object

    Parameters
    ----------
    qpi: QPImage
        Quantitative phase image data
    r0: float
        Approximate radius of the sphere [m]
    method: str
        The method used to determine the refractive index
        can either be "edge" (determine the radius from the
        edge detected in the phase image) or "image" (perform
        a 2D phase image fit).
    model: str
        The light-scattering model used by `method`. If
        `method` is "edge", only "projection" is allowed.
        If `method` is "image", `model` can be one of
        "mie", "projection", "rytov", or "rytov-sc".
    edgekw: dict
        Keyword arguments for tuning the edge detection algorithm,
        see :func:`qpsphere.edgefit.contour_canny`.
    imagekw: dict
        Keyword arguments for tuning the image fitting algorithm,
        see :func:`qpsphere.imagefit.alg.match_phase`
    ret_center: bool
        If True, return the center coordinate of the sphere.
    ret_pha_offset: bool
        If True, return the phase image background offset.
    ret_qpi: bool
        If True, return the modeled data as a :class:`qpimage.QPImage`.

    Returns
    -------
    n: float
        Computed refractive index
    r: float
        Computed radius [m]
    c: tuple of floats
        Only returned if `ret_center` is True
        Center position of the sphere [px]
    pha_offset: float
        Only returned if `ret_pha_offset` is True
        Phase image background offset
    qpi_sim: qpimage.QPImage
        Only returned if `ret_qpi` is True
        Modeled data

    Notes
    -----
    If `method` is "image", then the "edge" method is used
    as a first step to estimate initial parameters for radius,
    refractive index, and position of the sphere using `edgekw`.
    If this behavior is not desired, please make use of the
    method :func:`qpsphere.imagefit.analyze`.
    """
    if method == "edge":
        if model != "projection":
            raise ValueError("`method='edge'` requires `model='projection'`!")
        n, r, c = edgefit.analyze(qpi=qpi,
                                  r0=r0,
                                  edgekw=edgekw,
                                  ret_center=True,
                                  ret_edge=False,
                                  )
        res = [n, r]
        if ret_center:
            res.append(c)
        if ret_pha_offset:
            res.append(0)
        if ret_qpi:
            qpi_sim = simulate(radius=r,
                               sphere_index=n,
                               medium_index=qpi["medium index"],
                               wavelength=qpi["wavelength"],
                               grid_size=qpi.shape,
                               model="projection",
                               pixel_size=qpi["pixel size"],
                               center=c)
            res.append(qpi_sim)
    elif method == "image":
        n0, r0, c0 = edgefit.analyze(qpi=qpi,
                                     r0=r0,
                                     edgekw=edgekw,
                                     ret_center=True,
                                     ret_edge=False,
                                     )
        res = imagefit.analyze(qpi=qpi,
                               model=model,
                               n0=n0,
                               r0=r0,
                               c0=c0,
                               imagekw=imagekw,
                               ret_center=ret_center,
                               ret_pha_offset=ret_pha_offset,
                               ret_qpi=ret_qpi
                               )
    else:
        raise NotImplementedError("`method` must be 'edge' or 'image'!")

    return res



import os
import warnings

import matplotlib.pylab as plt
import numpy as np

import qpimage

from .interp import SpherePhaseInterpolator


def match_phase(qpi, model, n0, r0, c0=None, pha_offset=0,
                fix_pha_offset=False, nrel=.10, rrel=.05, crel=.05,
                stop_dn=.0005, stop_dr=.0010, stop_dc=1, min_iter=3,
                max_iter=100, ret_center=False, ret_pha_offset=False,
                ret_qpi=False, ret_num_iter=False, ret_interim=False,
                verbose=0, verbose_out_prefix="./verbose_out/field"
                ):
    """Fit a scattering model to a quantitative phase image

    Parameters
    ----------
    qpi: qpimage.QPImage
        QPI data to fit (e.g. experimental data)
    model: str
        Name of the light-scattering model
        (see :const:`qpsphere.models.available`)
    n0: float
        Initial refractive index of the sphere
    r0: float
        Initial radius of the sphere [m]
    c0: tuple of (float, float)
        Initial center position of the sphere in ndarray index
        coordinates [px]; if set to `None` (default), the center
        of the image is used.
    pha_offset: float
        Initial phase offset [rad]
    fix_pha_offset: bool
        If True, do not fit the phase offset `pha_offset`. The phase
        offset is determined from the mean of all pixels whose absolute
        phase is

        - below 1% of the modeled phase and
        - within a 5px or 20% border (depending on which is larger)
          around the phase image.
    nrel: float
        Determines the border of the interpolation range for the
        refractive index: [n-(n-nmed)*nrel, n+(n-nmed)*nrel]
        with nmed=qpi["medium_index"] and, initially, n=n0.
    rrel: float
        Determines the border of the interpolation range for the
        radius: [r*(1-rrel), r*(1+rrel)] with, initially, r=r0.
    crel: float
        Determines the border of the interpolation range for the
        center position: [cxy - dc, cxy + dc] with the center
        position (along x or y) cxy, and the interval radius dc
        defined by dc=max(lambda, crel * r0) with the vacuum
        wavelength lambda=qpi["wavelenght"].
    stop_dn: float
        Stopping criterion for refractive index
    stop_dr: float
        Stopping criterion for radius
    stop_dc: float
        Stopping criterion for lateral offsets
    min_iter: int
        Minimum number of fitting iterations to perform
    max_iter: int
        Maximum number of fitting iterations to perform
    ret_center: bool
        If True, return the fitted center coordinates
    ret_pha_offset: bool
        If True, return the fitted phase offset
    ret_qpi: bool
        If True, return the final fit as a data set
    ret_num_iter: bool
        If True, return the number of iterations
    ret_interim: bool
        If True, return intermediate parameters of each iteration
    verbose: int
        Higher values increase verbosity
    verbose_out_prefix: str
        Path to where images are saved at verbosity levels > 1

    Returns
    -------
    n: float
        Fitted refractive index
    r: float
        Fitted radius [m]
    c: tuple of (float, float)
        Only returned if `ret_center` is True
        Center position of the sphere in ndarray index coordinates [px]
    pha_offset: float
        Only returned if `ret_pha_offset` is True
        Fitted phase offset [rad]
    qpi: qpimage.QPImage
        Only returned if `ret_qpi` is True
        Simulation using `model` with the final fit parameters
    num_iter: int
        Only returned if `ret_num_iter` is True
        Number of iterations performed; negative number is
        returned when iteration fails
    interim: list
        Only returned if `ret_interim` is True
        Intermediate fitting parameters
    """
    if not isinstance(qpi, qpimage.QPImage):
        raise ValueError("`qpi` must be instance of `QPImage`!")
    for var in ["medium index", "pixel size", "wavelength"]:
        if var not in qpi:
            raise ValueError("meta data '{}' not defined in `qpi`!")

    if c0 is None:
        c0 = [qpi.shape[0] / 2, qpi.shape[1] / 2]

    model_kwargs = {"radius": r0,
                    "sphere_index": n0,
                    "medium_index": qpi["medium index"],
                    "wavelength": qpi["wavelength"],
                    "pixel_size": qpi["pixel size"],
                    "grid_size": qpi.shape,
                    "center": c0
                    }

    spi = SpherePhaseInterpolator(model=model,
                                  model_kwargs=model_kwargs,
                                  pha_offset=pha_offset,
                                  nrel=nrel,
                                  rrel=rrel,
                                  verbose=verbose)

    # Results recorder to detect stuck iterations
    recorder = []

    # intermediate results
    interim = []
    interim.append([0, spi.params])

    phase = qpi.pha
    range_ipol = 47
    range_off = 13
    # allow to vary center offset for 5 % of radius or 1 wavelengths
    dc = max(qpi["wavelength"], crel * r0) / qpi["pixel size"]  # [px]
    if verbose:
        print("Starting phase fitting.")
    ii = 0
    message = None
    while True:
        if verbose > 1:
            mphase = spi.get_phase()
            plot_phase_errors(phase, mphase, n0, r0,
                              spi.params, ii, model=model,
                              verbose_out_prefix=verbose_out_prefix)

        ii += 1

        # remember old values
        r_old = spi.radius
        n_old = spi.sphere_index

        # 1st step: vary radius
        rs = np.linspace(
            spi.range_r[0], spi.range_r[1], range_ipol, endpoint=True)
        assert np.allclose(np.min(np.abs(rs - spi.radius)), 0)
        lsqs = []
        for ri in rs:
            phasei = spi.get_phase(rintp=ri)
            lsqs.append(sq_phase_diff(phase, phasei))
        idr = np.argmin(lsqs)
        spi.radius = rs[idr]

        # 2nd step: vary n_object
        ns = np.linspace(
            spi.range_n[0], spi.range_n[1], range_ipol, endpoint=True)
        assert np.allclose(np.min(np.abs(ns - spi.sphere_index)), 0)
        lsqs = []
        for ni in ns:
            phasei = spi.get_phase(nintp=ni)
            lsqs.append(sq_phase_diff(phase, phasei))
        idn = np.argmin(lsqs)
        spi.sphere_index = ns[idn]

        # 3rd step: vary center position
        x = np.linspace(-dc, dc, range_off, endpoint=True)
        assert np.allclose(np.min(np.abs(x)), 0)
        xintp, yintp = np.meshgrid(x, x)
        lsqs = []
        for xoff, yoff in zip(xintp.flatten(), yintp.flatten()):
            phasei = spi.get_phase(delta_offset_x=xoff, delta_offset_y=yoff)
            err = sq_phase_diff(phase, phasei)
            lsqs.append(err)

        idc = np.argmin(lsqs)
        deltax = xintp.flatten()[idc]
        deltay = yintp.flatten()[idc]

        # offsets must be added incrementally, because they are not overridden
        # in the 3rd step
        spi.posx_offset = spi.posx_offset - deltax
        spi.posy_offset = spi.posy_offset - deltay

        if not fix_pha_offset:
            # Use average phase at image border without sphere
            cabphase = spi.get_phase() - spi.pha_offset
            # Determine background
            cabphase[np.abs(cabphase) > .01 * np.abs(cabphase).max()] = np.nan
            cb_border = max(5, min(cabphase.shape) // 5)
            cabphase[cb_border:-cb_border, cb_border:-cb_border] = np.nan
            phai_offset = np.nanmean(cabphase - phase)

            if np.isnan(phai_offset):
                phai_offset = 0

            spi.pha_offset = - phai_offset

        if verbose == 1:
            print("Iteration {}: n={:.5e}, r={:.5e}m".format(ii,
                                                             spi.sphere_index,
                                                             spi.radius))
        elif verbose > 1:
            print("Iteration {}: {}", ii, spi.params)

        interim.append([ii, spi.params])

        # update accuracies
        if (idn > range_ipol / 2 - range_ipol / 10 and
                idn < range_ipol / 2 + range_ipol / 10):
            spi.dn /= 2
            if verbose > 1:
                print("Halved search interval: spi.dn={:.8f}".format(spi.dn))
        if (idr > range_ipol / 2 - range_ipol / 10 and
                idr < range_ipol / 2 + range_ipol / 10):
            spi.dr /= 2
            if verbose > 1:
                print("Halved search interval: spi.dr={:.8f}".format(spi.dr))
        if deltax**2 + deltay**2 < dc**2:
            dc /= 2
            if verbose > 1:
                print("Halved search interval: dc={:.8f}".format(dc))

        if ii < min_iter:
            if verbose:
                print("Keep iterating because `min_iter`={}.".format(min_iter))
            continue
        elif ii > max_iter:
            ii *= -1
            if verbose:
                print("Stopping iteration: reached `max_iter`={}".format(
                    max_iter))
            message = "fail, reached maximum number of iterations"
            break

        if stop_dc:
            # check movement of center location and enforce next iteration
            curoff = np.sqrt(deltax**2 + deltay**2)
            if curoff > stop_dc:
                if verbose:
                    print("Keep iterating because center location moved by "
                          + "{} > `stop_dc`={}.".format(curoff, stop_dc))
                continue

        if (abs(spi.radius - r_old) / spi.radius < stop_dr and
                abs(spi.sphere_index - n_old) < stop_dn):
            # Radius, refractive index, and center position changed below
            # user-defined threshold.
            if verbose:
                print("Stopping iteration: `stop_dr` and `stop_dn` satisfied")
            message = "success, satisfied stopping criteria"
            break

        thisresult = (spi.sphere_index, spi.radius)
        recorder.append(thisresult)
        if recorder.count(thisresult) > 2:
            ii *= -1
            # We have already had this result 2 times and therefore we abort.
            # TODO:
            # - Select the one with the least error
            warnings.warn("Aborting stuck iteration for {}!".format(qpi))
            if verbose:
                print("Stop iteration: encountered same parameters twice.")
            message = "fail, same parameters encountered twice"
            break

        if verbose > 1:
            infostring = ""
            if not abs(spi.sphere_index - n_old) < stop_dn:
                infostring += " delta_n = {} > {}".format(
                    abs(spi.sphere_index - n_old), stop_dn)
            if not abs(spi.radius - r_old) / spi.radius < stop_dr:
                infostring += " delta_r = {} > {}".format(
                    abs(spi.radius - r_old) / spi.radius, stop_dr)
            print("Keep iterating: {} (no convergence)".format(infostring))

    if verbose:
        print("Number of iterations: {}".format(ii))
        print("Stopping rationale: {}".format(message))

    if verbose > 1:
        mphase = spi.get_phase()
        plot_phase_errors(phase, mphase, n0, r0,
                          spi.params, ii, model=model,
                          verbose_out_prefix=verbose_out_prefix)

    res = [spi.sphere_index, spi.radius]

    if ret_center:
        res += [[spi.posx_offset, spi.posy_offset]]
    if ret_pha_offset:
        res += [spi.pha_offset]
    if ret_qpi:
        res += [spi.compute_qpi()]
    if ret_num_iter:
        res += [ii]
    if ret_interim:
        res += [interim]

    return res


def sq_phase_diff(pha_a, pha_b):
    """Compute sum of squares error between two arrays

    Parameters
    ----------
    pha_a, pha_b: 2d real-valued np.ndarrays
        Phase data to compare

    Returns
    -------
    sumsq: float
        Sum of squares of differences
    """
    err = np.sum((pha_a - pha_b)**2)
    return err


def plot_phase_errors(phase, mphase, n0, r0, spi_params, ii,
                      model, verbose_out_prefix):
    """Output phase image error as PNG and TXT files

    Parameters
    ----------
    phase: 2d real-valued np.ndarray
        phase image
    mphase: 2d real-valued np.ndarray
        reference phase image
    n0: float
        initial object index
    r0: float
        initial object radius [m]
    spi_params: dict
        parameter dictionary of :func:`SpherePhaseInterpolator`
    ii: int
        iteration index
    model: str
        sphere model name
    verbose_out_prefix: str
        path for filename prefix to save PNG and TXT files to.
        Image file names are formatted as:
        `{verbose_out_prefix}_phasematch_iter_{ii}_{model}.png`.
        Text file names are formatted as:
        `{verbose_out_prefix}_trace_{model}.txt`.
    """
    n = spi_params["sphere_index"]
    r = spi_params["radius"]

    phasekwargs = {"vmin": np.min(phase),
                   "vmax": np.max(phase),
                   "interpolation": "nearest"}
    errkwargs = {"vmin": -np.max(phase) / 5,
                 "vmax": np.max(phase) / 5,
                 "interpolation": "nearest",
                 "cmap": "coolwarm"}
    txtkwargs = {"verticalalignment": "top",
                 "horizontalalignment": "left",
                 "color": "white",
                 "fontsize": "15"}

    _fig, axes = plt.subplots(1, 3, figsize=(12, 4))
    axes = axes.flatten()

    ma1 = axes[0].imshow(phase, **phasekwargs)
    axes[0].text(0, 0, "n={:.5f}\nr={:.5f}um".format(
        n0, r0 * 1e6), **txtkwargs)

    ma2 = axes[1].imshow(mphase, **phasekwargs)
    axes[1].text(0, 0, "n={:.5f}\nr={:.5f}um".format(n, r * 1e6), **txtkwargs)

    ma3 = axes[2].imshow(phase - mphase, **errkwargs)

    # titles
    axes[0].set_title("original phase [rad]")
    axes[1].set_title("{} phase iter{} [rad]".format(model, ii))
    axes[2].set_title("difference iter{} [rad]".format(ii))

    # color bars
    plt.colorbar(ma1, ax=axes[0], fraction=.045, pad=.01)
    plt.colorbar(ma2, ax=axes[1], fraction=.045, pad=.01)
    plt.colorbar(ma3, ax=axes[2], fraction=.045, pad=.01)

    plt.tight_layout()
    outpath = verbose_out_prefix + \
        "phasematch_iter_{:04d}_{}.png".format(ii, model)
    if not os.path.exists(os.path.dirname(outpath)):
        os.mkdir(os.path.dirname(outpath))
    plt.savefig(outpath)

    plt.close()

    # write trace
    trout = verbose_out_prefix + "trace_{}.txt".format(model)
    with open(trout, "a") as fd:
        parms = [n0, r0, n, r, ii]
        parms = ["{:.10e}".format(p) for p in parms]
        fd.write(" ".join(parms) + "\n")



import numpy as np
import scipy.special as spspec
import scipy.interpolate as spinterp
from skimage.restoration import unwrap

import qpimage


def interpolate_grid(cin, cout, data, fillval=0):
    """Return the 2D field averaged radially w.r.t. the center"""
    if np.iscomplexobj(data):
        phase = interpolate_grid(
            cin, cout, unwrap.unwrap_phase(np.angle(data)), fillval=0)
        ampli = interpolate_grid(cin, cout, np.abs(data), fillval=1)
        return ampli * np.exp(1j * phase)

    ipol = spinterp.interp2d(x=cin[0], y=cin[1], z=data,
                             kind="linear",
                             copy=False,
                             bounds_error=False,
                             fill_value=fillval)
    return ipol(cout[0], cout[1])


def rytov(radius=5e-6, sphere_index=1.339, medium_index=1.333,
          wavelength=550e-9, pixel_size=1e-7, grid_size=(80, 80),
          center=(39.5, 39.5), radius_sampling=42):
    """Field behind a dielectric sphere in the Rytov approximation

    Parameters
    ----------
    radius: float
        Radius of the sphere [m]
    sphere_index: float
        Refractive index of the sphere
    medium_index: float
        Refractive index of the surrounding medium
    wavelength: float
        Vacuum wavelength of the imaging light [m]
    pixel_size: float
        Pixel size [m]
    grid_size: tuple of floats
        Resulting image size in x and y [px]
    center: tuple of floats
        Center position in image coordinates [px]
    radius_sampling: int
        Number of pixels used to sample the sphere radius when
        computing the Rytov field. The default value of 42
        pixels is a reasonable number for single-cell analysis.

    Returns
    -------
    qpi: qpimage.QPImage
        Quantitative phase data set
    """
    # sample the sphere radius with approximately 42px
    # (rounded to next integer pixel size)
    samp_mult = radius_sampling * pixel_size / radius
    sizex = grid_size[0] * samp_mult
    sizey = grid_size[1] * samp_mult
    grid_size_sim = [np.int(np.round(sizex)),
                     np.int(np.round(sizey))]
    size_factor = grid_size_sim[0] / grid_size[0]
    pixel_size_sim = pixel_size / size_factor

    field = sphere_prop_fslice_bessel(radius=radius,
                                      sphere_index=sphere_index,
                                      medium_index=medium_index,
                                      wavelength=wavelength,
                                      pixel_size=pixel_size_sim,
                                      grid_size=grid_size_sim,
                                      lD=0,
                                      approx="rytov",
                                      zeropad=5,
                                      oversample=1
                                      )

    # interpolate field back to original image coordinates
    # simulation coordinates
    x_sim = (np.arange(grid_size_sim[0]) + .5) * pixel_size_sim
    y_sim = (np.arange(grid_size_sim[1]) + .5) * pixel_size_sim
    # output coordinates
    x = (np.arange(grid_size[0]) + grid_size[0] / 2 - center[0]) * pixel_size
    y = (np.arange(grid_size[1]) + grid_size[1] / 2 - center[1]) * pixel_size

    # Interpolate resulting field with original extent
    field = interpolate_grid(cin=(y_sim, x_sim),
                             cout=(y, x),
                             data=field)

    meta_data = {"pixel size": pixel_size,
                 "wavelength": wavelength,
                 "medium index": medium_index,
                 "sim center": center,
                 "sim radius": radius,
                 "sim index": sphere_index,
                 }
    qpi = qpimage.QPImage(data=field, which_data="field",
                          meta_data=meta_data)
    return qpi


def sphere_prop_fslice_bessel(radius, sphere_index, medium_index,
                              wavelength=550e-9, pixel_size=1e-7,
                              grid_size=(80, 80), lD=0, approx="rytov",
                              zeropad=5, oversample=1
                              ):
    """Compute the projection of a disc using the Fourier slice theorem
    and the Bessel function of the first kind of order 1.

    Parameters
    ----------
    radius: float
        Radius of the sphere [m]
    sphere_index: float
        Refractive index of the sphere
    medium_index: float
        Refractive index of the surrounding medium
    wavelength: float
        Vacuum wavelength of the imaging light [m]
    pixel_size: float
        Pixel size [m]
    grid_size: tuple of floats
        Resulting image size in x and y [px]
    center: tuple of floats
        Center position in image coordinates [px]
    lD: float
        The axial distance [m] from the center of the sphere
        at which the field is computed.
    approx: str
        Which approximation to use (either "born" or "rytov")
    zeropad: int
        Zero-padding factor
    oversample: int
        Oversampling factor
    """
    assert approx in ["born", "rytov"]
    assert oversample > 0
    assert zeropad > 0
    assert isinstance(oversample, int)
    assert isinstance(zeropad, int)

    # convert everything to pixels
    radius /= pixel_size
    wavelength /= pixel_size
    lD /= pixel_size

    # apply over-sampling and zero-padding
    wavelength *= oversample
    opad_size = np.array(grid_size) * zeropad * oversample

    assert (int(s) == s for s in opad_size), "grid_size must be integer type"
    opad_size = np.array(np.round(opad_size), dtype=int)
    grid_size = np.array(np.round(grid_size), dtype=int)

    kx = 2 * np.pi * \
        np.fft.ifftshift(np.fft.fftfreq(opad_size[0])).reshape(-1, 1)
    ky = 2 * np.pi * \
        np.fft.ifftshift(np.fft.fftfreq(opad_size[1])).reshape(1, -1)
    km = 2 * np.pi * medium_index / wavelength

    filter_klp = (kx**2 + ky**2 < km**2)

    kz = np.sqrt((km**2 - kx**2 - ky**2) * filter_klp) - km

    r = np.sqrt((kx**2 + ky**2 + kz**2) * filter_klp) / (2 * np.pi)

    comp_id = r != 0
    F = np.zeros_like(r)
    F[comp_id] = spspec.spherical_jn(1, r[comp_id] * radius * np.pi * 2) \
        * radius**2 / r[comp_id] * 2
    # center has analytical value
    center_fft = np.where(np.abs(kx) + np.abs(ky) + np.abs(kz) == 0)
    F[center_fft] = 4 / 3 * np.pi * radius**3

    # object amplitude
    F *= km**2 * ((sphere_index / medium_index)**2 - 1)

    # prefactor A
    M = 1. / km * np.sqrt((km**2 - kx**2 - ky**2) * filter_klp)

    # division factor
    A = -2j * km * M * np.exp(-1j * km * M * lD)

    # rotate phase by half a pixel so the ifft is centered in real space
    if grid_size[0] % 2:
        doffx = 0
    else:
        doffx = .5
    if grid_size[1] % 2:
        doffy = 0
    else:
        doffy = .5
    transl = np.exp(1j * ((doffx) * kx + (doffy) * ky))

    valid = F != 0
    Fconv = np.zeros((opad_size[0], opad_size[1]), dtype=complex)
    Fconv[valid] = F[valid] / A[valid] * transl[valid]

    p = np.fft.ifftn(np.fft.fftshift(Fconv))

    p = np.fft.ifftshift(p)

    if oversample > 1:
        p = p[::oversample, ::oversample]

    if zeropad > 1:
        # Slice
        a0, a1 = np.array(np.floor(opad_size / 2), dtype=int) // oversample
        b0, b1 = np.array(np.floor(grid_size / 2), dtype=int)
        if grid_size[0] % 2 != 0:
            of0 = 1
            a0 += 1
        else:
            of0 = 0
        if grid_size[1] % 2 != 0:
            of1 = 1
            a1 += 1
        else:
            of1 = 0
        # remove zero-padding
        p = p[a0 - b0:a0 + b0 + of0, a1 - b1:a1 + b1 + of1]

    if approx == "born":
        # norm = (u0 + ub)/u0
        # norm = 1 + ub/u0
        return 1 + p / np.exp(1j * km * lD)
    elif approx == "rytov":
        # norm = (u0 + ur)/u0
        # ur = u0 ( exp(ub/u0) -1 )
        # norm = ( u0 + u0 *(exp(ub/u0)-1) )/u0
        # norm = exp(ub/u0)
        return np.exp(p / np.exp(1j * km * lD))



"""Comparison of light-scattering models

The phase error map allows a comparison of the ability of
the modeling methods implemented in qpsphere to reproduce
the phase delay introduced by a dielectric sphere.
For a quantitative comparison, see reference :cite:`Mueller2018`.
"""
import matplotlib.pylab as plt
import qpsphere

kwargs = {"radius": 10e-6,  # 10µm
          "sphere_index": 1.380,  # cell
          "medium_index": 1.335,  # PBS
          "wavelength": 647.1e-9,  # krypton laser
          "grid_size": (200, 200),
          }

px_size = 3 * kwargs["radius"] / kwargs["grid_size"][0]
kwargs["pixel_size"] = px_size

# mie (long computation time)
qpi_mie = qpsphere.simulate(model="mie", **kwargs)

# mie averaged
qpi_mie_avg = qpsphere.simulate(model="mie-avg", **kwargs)

# rytov corrected
qpi_ryt_sc = qpsphere.simulate(model="rytov-sc", **kwargs)

# rytov
qpi_ryt = qpsphere.simulate(model="rytov", **kwargs)

# projection
qpi_proj = qpsphere.simulate(model="projection", **kwargs)

kwargs = {"vmin": -.5,
          "vmax": .5,
          "cmap": "seismic"}

plt.figure(figsize=(8, 6.8))

ax1 = plt.subplot(221, title="Mie (averaged)")
pmap = plt.imshow(qpi_mie.pha - qpi_mie_avg.pha, **kwargs)

ax2 = plt.subplot(222, title="Rytov (corrected)")
plt.imshow(qpi_mie.pha - qpi_ryt_sc.pha, **kwargs)

ax3 = plt.subplot(223, title="Rytov")
plt.imshow(qpi_mie.pha - qpi_ryt.pha, **kwargs)

ax4 = plt.subplot(224, title="projection")
plt.imshow(qpi_mie.pha - qpi_proj.pha, **kwargs)

# disable axes
for ax in [ax1, ax2, ax3, ax4]:
    ax.axis("off")
    plt.colorbar(pmap, ax=ax, fraction=.045, pad=0.04,
                 label="phase error [rad]")

plt.tight_layout(w_pad=0, h_pad=0)
plt.show()



import nrefocus
import numpy as np
import qpimage

from ._bhfield import simulate_sphere


def mie(radius=5e-6, sphere_index=1.339, medium_index=1.333,
        wavelength=550e-9, pixel_size=1e-7, grid_size=(80, 80),
        center=(39.5, 39.5), arp=True):
    """Mie-simulated field behind a dielectric sphere

    Parameters
    ----------
    radius: float
        Radius of the sphere [m]
    sphere_index: float
        Refractive index of the sphere
    medium_index: float
        Refractive index of the surrounding medium
    wavelength: float
        Vacuum wavelength of the imaging light [m]
    pixel_size: float
        Pixel size [m]
    grid_size: tuple of floats
        Resulting image size in x and y [px]
    center: tuple of floats
        Center position in image coordinates [px]
    arp: bool
        Use arbitrary precision (ARPREC) in BHFIELD computations

    Returns
    -------
    qpi: qpimage.QPImage
        Quantitative phase data set
    """
    # simulation parameters
    radius_um = radius * 1e6  # radius of sphere in um
    propd_um = radius_um  # simulate propagation through full sphere
    propd_lamd = radius / wavelength  # radius in wavelengths
    wave_nm = wavelength * 1e9
    # Qpsphere models define the position of the sphere with an index in
    # the array (because it is easier to work with). The pixel
    # indices run from (0, 0) to grid_size (without endpoint). BHFIELD
    # requires the extent to be given in µm. The distance in µm between
    # first and last pixel (measured from pixel center) is
    # (grid_size - 1) * pixel_size,
    size_um = (np.array(grid_size) - 1) * pixel_size * 1e6
    # The same holds for the offset. If we use size_um here,
    # we already take into account the half-pixel offset.
    offset_um = np.array(center) * pixel_size * 1e6 - size_um / 2

    kwargs = {"radius_sphere_um": radius_um,
              "refractive_index_medium": medium_index,
              "refractive_index_sphere": sphere_index,
              "measurement_position_um": propd_um,
              "wavelength_nm": wave_nm,
              "size_simulation_um": size_um,
              "shape_grid": grid_size,
              "offset_x_um": offset_um[0],
              "offset_y_um": offset_um[1]}

    background = np.exp(1j * 2 * np.pi * propd_lamd * medium_index)

    field = simulate_sphere(arp=arp, **kwargs) / background

    # refocus
    refoc = nrefocus.refocus(field,
                             d=-(radius / pixel_size),
                             nm=medium_index,
                             res=wavelength / pixel_size)

    meta_data = {"pixel size": pixel_size,
                 "wavelength": wavelength,
                 "medium index": medium_index,
                 "sim center": center,
                 "sim radius": radius,
                 "sim index": sphere_index,
                 }

    qpi = qpimage.QPImage(data=refoc, which_data="field",
                          meta_data=meta_data)
    return qpi



import numpy as np

from qpsphere.models import projection


def test_basic():
    data = np.array([0.6854384, 0.62821467, 0.41126304,
                     0.62821467, 0.56522698, 0.30653737,
                     0.41126304, 0.30653737, 0.])

    qpi = projection(grid_size=(3, 3),
                     center=(0, 0),
                     pixel_size=2e-6)
    assert np.allclose(data, qpi.pha.flatten())


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import numpy as np

from qpsphere.models import mod_rytov_sc


def test_parameter_inversion():
    n, r = 1.45, 5.6e-6
    nmed = 1.333
    r_sc, n_sc = mod_rytov_sc.correct_rytov_output(radius=r,
                                                   sphere_index=n,
                                                   medium_index=1.333,
                                                   radius_sampling=42)
    assert r_sc < r
    assert n_sc > n

    r2, n2 = mod_rytov_sc.correct_rytov_sc_input(radius_sc=r_sc,
                                                 sphere_index_sc=n_sc,
                                                 medium_index=nmed,
                                                 radius_sampling=42)

    assert np.allclose(r, r2, atol=1e-15, rtol=0)
    assert np.allclose(n, n2, atol=1e-15, rtol=0)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



"""Wrapper around BHFIELD binaries for Mie field computation

see https://github.com/RI-imaging/QPI-binaries/tree/master/BHFIELD
"""
from . import fetch  # noqa: F401
from .wrap import simulate_sphere  # noqa: F401



import warnings

import numpy as np

from qpsphere.models import _bhfield as bh


def test_default_simulation():
    data = np.array([0.99915 + 0.0059473j, 1.01950 - 0.0067643j,
                     1.01950 - 0.0067643j, 0.99915 + 0.0059473j,
                     1.02340 - 0.0037607j, 0.97630 + 0.4906j,
                     0.97630 + 0.4906j, 1.02340 - 0.0037607j,
                     1.02340 - 0.0037607j, 0.97630 + 0.4906j,
                     0.97630 + 0.4906j, 1.02340 - 0.0037607j,
                     0.99915 + 0.0059473j, 1.01950 - 0.0067643j,
                     1.01950 - 0.0067643j, 0.99915 + 0.0059473j])

    field = bh.simulate_sphere(shape_grid=(4, 4))
    assert np.allclose(field.flatten(), data)


def test_force_arp_warning():
    with warnings.catch_warnings(record=True) as rw:
        bh.simulate_sphere(radius_sphere_um=7,
                           refractive_index_sphere=1.5,
                           size_simulation_um=(20, 20),
                           shape_grid=(4, 4),
                           arp=False)
        msg = str(rw[0].message).lower()
        assert msg.count("bhfield")
        assert msg.count("standard precision failed")
        assert msg.count("retrying with arbitrary precision")


def test_get_binary():
    path1 = bh.fetch.get_binary(arp=False)
    path2 = bh.fetch.get_binary(arp=True)
    assert path1.exists()
    assert path2.exists()


def test_known_fail():
    try:
        bh.simulate_sphere(radius_sphere_um=50,
                           size_simulation_um=(70, 70),
                           shape_grid=(4, 4))
    except bh.wrap.BHFIELDExecutionError:
        pass
    else:
        assert False, "This simulation should not work with BHFIELD"


def test_shape():
    f1 = bh.simulate_sphere(shape_grid=(1, 4))
    f2 = bh.simulate_sphere(shape_grid=(4, 1))
    assert f1.shape == (1, 4)
    assert f2.shape == (4, 1)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import numpy as np

from . import mod_rytov


#: correction parameters (see :func:`correct_rytov_output`)
RSC_PARAMS = {42: {"na": 1.936,
                   "nb": -0.012,
                   "ra": -2.431,
                   "rb": -.753,
                   "rc": 1.001}
              }


def correct_rytov_sc_input(radius_sc, sphere_index_sc, medium_index,
                           radius_sampling):
    """Inverse correction of refractive index and radius for Rytov

    This method returns the inverse of :func:`correct_rytov_output`.

    Parameters
    ----------
    radius_sc: float
        Systematically corrected radius of the sphere [m]
    sphere_index_sc: float
        Systematically corrected refractive index of the sphere
    medium_index: float
        Refractive index of the surrounding medium
    radius_sampling: int
        Number of pixels used to sample the sphere radius when
        computing the Rytov field.

    Returns
    -------
    radius: float
        Fitted radius of the sphere [m]
    sphere_index: float
        Fitted refractive index of the sphere

    See Also
    --------
    correct_rytov_output: the inverse of this method
    """
    params = get_params(radius_sampling)

    # sage script:
    # var('sphere_index, sphere_index_sc, na, nb, medium_index')
    # x = sphere_index / medium_index - 1
    # eq = sphere_index_sc == sphere_index + ( na*x^2 + nb*x) * medium_index
    # solve([eq], [sphere_index])
    # (take the positive sign solution)
    na = params["na"]
    nb = params["nb"]

    prefac = medium_index / (2 * na)
    sm = 2 * na - nb - 1
    rt = nb**2 - 4 * na + 2 * nb + 1 + 4 / medium_index * na * sphere_index_sc
    sphere_index = prefac * (sm + np.sqrt(rt))

    x = sphere_index / medium_index - 1

    radius = radius_sc / (params["ra"] * x**2
                          + params["rb"] * x
                          + params["rc"])

    return radius, sphere_index


def correct_rytov_output(radius, sphere_index, medium_index, radius_sampling):
    """Error-correction of refractive index and radius for Rytov

    This method corrects the fitting results for `radius`
    :math:`r_\text{Ryt}` and `sphere_index` :math:`n_\text{Ryt}`
    obtained using :func:`qpsphere.models.mod_rytov.rytov` using
    the approach described in :cite:`Mueller2018` (eqns. 3,4, and 5).

    .. math::

        n_\text{Ryt-SC} &= n_\text{Ryt} + n_\text{med} \cdot
                           \left( a_n x^2 + b_n x + c_n \right)
        r_\text{Ryt-SC} &= r_\text{Ryt} \cdot
                           \left( a_r x^2 +b_r x + c_r \right)
        &\text{with~} x = \frac{n_\text{Ryt}}{n_\text{med}} - 1

    Parameters
    ----------
    radius: float
        Fitted radius of the sphere :math:`r_\text{Ryt}` [m]
    sphere_index: float
        Fitted refractive index of the sphere :math:`n_\text{Ryt}`
    medium_index: float
        Refractive index of the surrounding medium :math:`n_\text{med}`
    radius_sampling: int
        Number of pixels used to sample the sphere radius when
        computing the Rytov field.

    Returns
    -------
    radius_sc: float
        Systematically corrected radius of the sphere
        :math:`r_\text{Ryt-SC}` [m]
    sphere_index_sc: float
        Systematically corrected refractive index of the sphere
        :math:`n_\text{Ryt-SC}`

    See Also
    --------
    correct_rytov_sc_input: the inverse of this method
    """
    params = get_params(radius_sampling)

    x = sphere_index / medium_index - 1

    radius_sc = radius * (params["ra"] * x**2
                          + params["rb"] * x
                          + params["rc"])

    sphere_index_sc = sphere_index + medium_index * (params["na"] * x**2
                                                     + params["nb"] * x)

    return radius_sc, sphere_index_sc


def get_params(radius_sampling):
    if radius_sampling not in RSC_PARAMS:
        msg = "`radius_sampling={}` not available. ".format(radius_sampling) \
              + "Please use one of: {}".format(RSC_PARAMS.keys())
        raise ValueError(msg)
    return RSC_PARAMS[radius_sampling]


def rytov_sc(radius=5e-6, sphere_index=1.339, medium_index=1.333,
             wavelength=550e-9, pixel_size=1e-7, grid_size=(80, 80),
             center=(39.5, 39.5), radius_sampling=42):
    """Field behind a dielectric sphere, systematically corrected Rytov

    Parameters
    ----------
    radius: float
        Radius of the sphere [m]
    sphere_index: float
        Refractive index of the sphere
    medium_index: float
        Refractive index of the surrounding medium
    wavelength: float
        Vacuum wavelength of the imaging light [m]
    pixel_size: float
        Pixel size [m]
    grid_size: tuple of floats
        Resulting image size in x and y [px]
    center: tuple of floats
        Center position in image coordinates [px]
    radius_sampling: int
        Number of pixels used to sample the sphere radius when
        computing the Rytov field. The default value of 42
        pixels is a reasonable number for single-cell analysis.

    Returns
    -------
    qpi: qpimage.QPImage
        Quantitative phase data set

    Notes
    -----

    """
    r_ryt, n_ryt = correct_rytov_sc_input(radius_sc=radius,
                                          sphere_index_sc=sphere_index,
                                          medium_index=medium_index,
                                          radius_sampling=radius_sampling)

    qpi = mod_rytov.rytov(radius=r_ryt,
                          sphere_index=n_ryt,
                          medium_index=medium_index,
                          wavelength=wavelength,
                          pixel_size=pixel_size,
                          grid_size=grid_size,
                          center=center,
                          radius_sampling=radius_sampling)

    # update correct simulation parameters
    qpi["sim radius"] = radius
    qpi["sim index"] = sphere_index

    return qpi



from .alg import match_phase


def analyze(qpi, model, n0, r0, c0=None, imagekw={},
            ret_center=False, ret_pha_offset=False, ret_qpi=False):
    """Fit refractive index and radius to a phase image of a sphere

    Parameters
    ----------
    qpi: QPImage
        Quantitative phase image information
    model: str
        Name of the light-scattering model
        (see :const:`qpsphere.models.available`)
    n0: float
        Approximate refractive index of the sphere
    r0: float
        Approximate radius of the sphere [m]
    c0: tuple of (float, float)
        Approximate center position in ndarray index coordinates [px];
        if set to `None` (default), the center of the image is used.
    imagekw: dict
        Additional keyword arguments to
        :func:`qpsphere.imagefit.alg.match_phase`.
    ret_center: bool
        Return the center coordinate of the sphere
    ret_pha_offset: bool
        If True, return the phase image background offset.
    ret_qpi: bool
        If True, return the modeled data as a :class:`qpimage.QPImage`.

    Returns
    -------
    n: float
        Computed refractive index
    r: float
        Computed radius [m]
    c: tuple of floats
        Only returned if `ret_center` is True
        Center position of the sphere [px]
    pha_offset: float
        Only returned if `ret_pha_offset` is True
        Phase image background offset
    qpi_sim: qpimage.QPImage
        Only returned if `ret_qpi` is True
        Modeled data
    """
    res = match_phase(qpi, model=model, n0=n0, r0=r0, c0=c0,
                      ret_center=ret_center,
                      ret_pha_offset=ret_pha_offset,
                      ret_qpi=ret_qpi,
                      **imagekw)
    return res



import pathlib
import shutil
import tempfile

from qpsphere import util


def copy_to_safe():
    """Move all binary files to a temporary folder"""
    # This does not require an internet connection if all
    # the files are already present in the cache or
    # resource directory.
    paths = util.download_binaries(package_dir=False)
    dtemp = tempfile.mkdtemp(prefix="qpsphere-resources_")
    for pp in paths:
        shutil.copy2(str(pp), dtemp)
    return pathlib.Path(dtemp)


def copy_restore(dtemp):
    """Restore all files to user cache dir and delete dtemp"""
    for pp in dtemp.iterdir():
        pp.rename(util.CACHE_PATH / pp.name)
    dtemp.rmdir()


def test_move_to_resources_and_priority():
    dtemp = copy_to_safe()
    # remove all files from package directory
    util.remove_binaries(package_dir=True)
    # test that resource directory contains no binaries
    nf = [pp for pp in util.RESCR_PATH.iterdir()]
    # only one file: "shipped_resources_go_here"
    assert len(nf) == 1
    # download binaries to package directory
    paths = util.download_binaries(package_dir=True)
    # test whether package directory contains binaries
    assert len(paths)
    # test whether binaries in package directroy get
    # highest priority
    paths2 = util.download_binaries(package_dir=False)
    for pp in paths2:
        assert pp == util.RESCR_PATH / pp.name
    # remove the files from the resource directory
    util.remove_binaries(package_dir=True)
    paths3 = util.download_binaries(package_dir=False)
    for pp in paths3:
        assert pp == util.CACHE_PATH / pp.name
    # populate resources dir and remove cache dir
    paths = util.download_binaries(package_dir=True)
    util.remove_binaries(package_dir=False)
    paths4 = util.download_binaries(package_dir=False)
    for pp in paths4:
        assert pp == util.RESCR_PATH / pp.name
    for pp in paths3:
        assert not pp.exists()
    # cleanup
    util.remove_binaries(package_dir=True)
    copy_restore(dtemp)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import nrefocus
import numpy as np
import scipy.interpolate as spinterp
from skimage.restoration import unwrap

import qpimage

from ._bhfield import simulate_sphere


def mie_avg(radius=5e-6, sphere_index=1.339, medium_index=1.333,
            wavelength=550e-9, pixel_size=1e-7, grid_size=(80, 80),
            center=(39.5, 39.5), interpolate=3, arp=True):
    """Mie-simulated non-polarized field behind a dielectric sphere

    Parameters
    ----------
    radius: float
        Radius of the sphere [m]
    sphere_index: float
        Refractive index of the sphere
    medium_index: float
        Refractive index of the surrounding medium
    wavelength: float
        Vacuum wavelength of the imaging light [m]
    pixel_size: float
        Pixel size [m]
    grid_size: tuple of floats
        Resulting image size in x and y [px]
    center: tuple of floats
        Center position in image coordinates [px]
    interpolate: int
        Compute the radial field with sampling that is by a factor of
        `interpolate` higher than the required data and interpolate the
        2D field from there.
    arp: bool
        Use arbitrary precision (ARPREC) in BHFIELD computations

    Returns
    -------
    qpi: qpimage.QPImage
        Quantitative phase data set
    """
    center = np.array(center)
    grid_size = np.array(grid_size)
    # simulation parameters
    radius_um = radius * 1e6  # radius of sphere in um
    propd_um = radius_um  # simulate propagation through full sphere
    propd_lamd = radius / wavelength  # radius in wavelengths
    wave_nm = wavelength * 1e9

    kwargs = {"radius_sphere_um": radius_um,
              "refractive_index_medium": medium_index,
              "refractive_index_sphere": sphere_index,
              "measurement_position_um": propd_um,
              "wavelength_nm": wave_nm}

    upres = wavelength / pixel_size * interpolate
    max_off = np.max(np.abs(grid_size / 2 - .5 - center))

    latsize = int(np.round((np.max(grid_size) + max_off)))

    num = latsize * upres / 2
    # find the maximum interpolation range in the 2d image

    bignum = int(np.ceil(np.sqrt(np.sum((np.array(grid_size) /
                                         2 + max_off)**2))) * interpolate)

    # Compare this number to the radius of the sphere and cut it off at
    # three times the radius.
    radnum = int(np.ceil(3 * radius / pixel_size) * interpolate)
    bignum = min(bignum, radnum)

    latsize *= bignum / num
    latsize *= wavelength * 1e6
    upres /= wavelength * 1e6

    background = np.exp(1j * 2 * np.pi * propd_lamd * medium_index)

    # [sic]: Not times upres
    ofx_px = grid_size[0] / 2 - center[0]
    ofy_px = grid_size[1] / 2 - center[1]

    kwargsx = kwargs.copy()
    kwargsx.update({"size_simulation_um": (latsize / 2, 1 / upres),
                    "shape_grid": (bignum, 1),
                    "offset_x_um": -latsize / 4,
                    "offset_y_um": 0})

    fieldx = simulate_sphere(arp=arp, **kwargsx) / background

    kwargsy = kwargs.copy()
    kwargsy.update({"size_simulation_um": (1 / upres, latsize / 2),
                    "shape_grid": (1, bignum),
                    "offset_x_um": 0,
                    "offset_y_um": -latsize / 4})
    fieldy = simulate_sphere(arp=arp, **kwargsy) / background

    field = (fieldx.flatten() + fieldy.flatten()) / 2

    xo = np.linspace(0, bignum, bignum, endpoint=True) / interpolate

    x = np.linspace(-grid_size[0] / 2,
                    grid_size[0] / 2,
                    grid_size[0] * interpolate,
                    endpoint=True)

    y = np.linspace(-grid_size[1] / 2,
                    grid_size[1] / 2,
                    grid_size[1] * interpolate,
                    endpoint=True)

    yv, xv = np.meshgrid(y, x)
    r = np.sqrt(xv**2 + yv**2)
    inpt_kwargs = {"kind": "linear",
                   "assume_sorted": True,
                   "bounds_error": False,
                   }

    ipltph = spinterp.interp1d(xo, np.unwrap(
        np.angle(field)), fill_value=0, **inpt_kwargs)
    ipltam = spinterp.interp1d(xo, np.abs(field), fill_value=1, **inpt_kwargs)
    phase2d = ipltph(r)
    field2d = ipltam(r) * np.exp(1j * phase2d)

    # Numerical refocusing
    # We need to perform numerical focusing with the upsampled array,
    # or else we will loose spatial information and the resulting
    # spherical image becomes asymmetric.
    refoc_field2d = nrefocus.refocus(field2d,
                                     d=-(radius / pixel_size) * interpolate,
                                     nm=medium_index,
                                     res=wavelength / pixel_size * interpolate)

    # Perform new interpolation on requested grid
    # We might introduce an offset here:
    phase = unwrap.unwrap_phase(np.angle(refoc_field2d), seed=47)

    # detect and remove multiple-2PI phase offset
    pha_offset = np.median(phase[:3])
    num_2pi = np.round(pha_offset / (2 * np.pi))
    if num_2pi:
        phase -= num_2pi * 2 * np.pi

    ampli = np.abs(refoc_field2d)
    intpp = spinterp.RectBivariateSpline(x, y, phase, kx=1, ky=1)
    intpa = spinterp.RectBivariateSpline(x, y, ampli, kx=1, ky=1)

    xp = np.linspace(-grid_size[0] / 2,
                     grid_size[0] / 2,
                     grid_size[0],
                     endpoint=False) + ofx_px

    yp = np.linspace(-grid_size[1] / 2,
                     grid_size[1] / 2,
                     grid_size[1],
                     endpoint=False) + ofy_px

    amp_off = intpa(xp, yp)
    pha_off = intpp(xp, yp)

    meta_data = {"pixel size": pixel_size,
                 "wavelength": wavelength,
                 "medium index": medium_index,
                 "sim center": center,
                 "sim radius": radius,
                 "sim index": sphere_index,
                 }

    qpi = qpimage.QPImage(data=(pha_off, amp_off),
                          which_data="phase,amplitude",
                          meta_data=meta_data)

    return qpi



"""Rytov-SC image fit applied to a Mie simulation

This examples illustrates how the refractive index and radius of
a sphere can be determined accurately using the 2D phase fitting
algorithm with the systematically corrected Rytov approximation.
"""
import matplotlib.pylab as plt

import qpsphere

# run simulation with averaged Mie model
r = 5e-6
n = 1.360
med = 1.333
c = (125, 133)
qpi = qpsphere.simulate(radius=r,
                        sphere_index=n,
                        medium_index=med,
                        wavelength=550e-9,
                        grid_size=(256, 256),
                        model="mie-avg",
                        center=c)

# Fitting Mie simulations with the systematically corrected Rytov
# approximation (`model="rytov sc"`) yields lower parameter errors
# compared to the non-corrected Rytov approximation (`model="rytov"`).
n_fit, r_fit, c_fit, qpi_fit = qpsphere.analyze(qpi=qpi,
                                                r0=4e-6,
                                                method="image",
                                                model="rytov-sc",
                                                imagekw={"verbose": 1},
                                                ret_center=True,
                                                ret_qpi=True)

# plot results
fig = plt.figure(figsize=(8, 3.5))
txtkwargs = {"verticalalignment": "bottom",
             "horizontalalignment": "right",
             "fontsize": 12}

ax1 = plt.subplot(121, title="ground truth (Mie theory)")
map1 = ax1.imshow(qpi.pha)
plt.colorbar(map1, ax=ax1, fraction=.046, pad=0.04, label="phase [rad]")
t1 = "n={:.3f}\nmed={:.3f}\nr={:.1f}µm\ncenter=({:d},{:d})".format(
    n, med, r * 1e6, c[0], c[1])
ax1.text(1, 0, t1, transform=ax1.transAxes, color="w", **txtkwargs)

ax2 = plt.subplot(122, title="Rytov-SC fit residuals")
map2 = ax2.imshow(qpi.pha - qpi_fit.pha, vmin=-.5, vmax=.5, cmap="seismic")
plt.colorbar(map2, ax=ax2, fraction=.046, pad=0.04, label="phase error [rad]")
t2 = "Δn/n={:.2e}\nΔr/r={:.2e}\nΔx={:.2e}\nΔy={:.2e}".format(
    abs(n - n_fit) / n, abs(r - r_fit) / r,
    c_fit[0] - c[0], c_fit[1] - c[1])
ax2.text(1, 0, t2, transform=ax2.transAxes, color="k", **txtkwargs)

plt.tight_layout()
plt.show()



"""Fetch BHFIELD binaries from GitHub"""
import contextlib
import hashlib
import pathlib
from pkg_resources import resource_filename
import platform
import stat
import urllib

import appdirs

CACHE_DIR = appdirs.AppDirs(appname="python-qpsphere").user_cache_dir
RESRC_DIR = resource_filename("qpsphere", "resources")


EXE_STD = "bhfield-std.exe"
EXE_ARP = "bhfield-arp.exe"

DL_REPO = "https://github.com/RI-imaging/QPI-binaries/raw/" \
          + "d2af4117917ade389958e105d36446830b404fe7" \
          + "/BHFIELD"

BHFIELD_MD5 = {"Windows": {
    "bhfield-std.exe": "673415a9d49cc4ae0bdcd2da7b78802a",
    "bhfield-arp.exe": "75d4b592a000944c947b7133996c9a4c",
},
    "Linux": {
    "bhfield-std.exe": "5141ca8101023b35288f1eeef9ab6ae5",
    "bhfield-arp.exe": "843a802dc822251f8a4b4505f2152619",
},
}


class BinaryNotAvailableError(BaseException):
    """Raised if the BHFIELD binary is not available for the current platform
    """
    pass


class BinaryMD5SumCheckError(BaseException):
    """Raised when the downloaded binary does not have the correct md5sum
    """
    pass


def download_binary(url, dest, md5, retries=3):
    print("qpsphere: Downloading '{}', please wait.".format(dest.name))
    for rr in range(retries):
        if dest.exists():
            dest.unlink()
        try:
            urlretrieve(url, dest)
        except BaseException:
            print("qpsphere: Download of {} failed ({}/{})".format(url,
                                                                  rr + 1,
                                                                  retries))
            continue
        else:
            break
    else:
        # Let the user see the error message
        urlretrieve(url, dest)

    # verify binary
    if md5sum(dest) != md5:
        msg = "MD5 sum of {} does not match {}!".format(dest, md5)
        raise BinaryMD5SumCheckError(msg)


def get_download_root():
    plat = platform.system()
    if plat == "Windows":
        path = "{}/{}".format(DL_REPO.strip("/"), "bin_win")
    elif plat == "Linux":
        path = "{}/{}".format(DL_REPO.strip("/"), "bin_linux")
    else:
        msg = "BHFIELD currently unavailable for platform '{}'!".format(plat)
        raise BinaryNotAvailableError(msg)
    return path


def get_binary(arp=False):
    plat = platform.system()
    dl_root = get_download_root()
    if arp:
        exe = EXE_ARP
    else:
        exe = EXE_STD

    # check if file is in resources directory
    res_path = pathlib.Path(RESRC_DIR) / exe
    if res_path.exists():
        ret_binary = res_path
    else:
        # download from GitHub
        ret_binary = pathlib.Path(CACHE_DIR) / exe
        url = "{}/{}".format(dl_root, exe)
        md5 = BHFIELD_MD5[plat][exe]

        if (not ret_binary.exists() or
                not md5sum(ret_binary) == md5):

            download_binary(url=url, dest=ret_binary, md5=md5)

    # make sure file is executable
    st = ret_binary.stat()
    ret_binary.chmod(st.st_mode | stat.S_IEXEC)
    return ret_binary


def get_binaries():
    """Download and return paths of all platform-specific binaries"""
    paths = []
    for arp in [False, True]:
        paths.append(get_binary(arp=arp))
    return paths


def md5sum(binary, blocksize=65536):
    hasher = hashlib.md5()
    with binary.open(mode="rb") as fd:
        buf = fd.read(blocksize)
        while len(buf) > 0:
            hasher.update(buf)
            buf = fd.read(blocksize)
    return hasher.hexdigest()


def urlretrieve(url, dest):
    # create destination directory
    dest.parent.mkdir(parents=True, exist_ok=True)
    # begin download
    with dest.open(mode="wb") as out_file:
        with contextlib.closing(urllib.request.urlopen(str(url))) as fp:
            block_size = 1024 * 8
            while True:
                block = fp.read(block_size)
                if not block:
                    break
                out_file.write(block)



import numpy as np

import qpsphere


def test_basic():
    r = 5e-6
    n = 1.339
    c = (11, 11)
    s = 25
    qpi = qpsphere.simulate(radius=r,
                            sphere_index=n,
                            medium_index=1.333,
                            wavelength=550e-9,
                            grid_size=(s, s),
                            model="projection",
                            pixel_size=3 * r / s,
                            center=c)

    # fit simulation with projection model
    n_fit, r_fit, c_fit, qpi_fit = qpsphere.analyze(qpi=qpi,
                                                    r0=r * .8,
                                                    method="image",
                                                    model="projection",
                                                    ret_center=True,
                                                    ret_qpi=True,
                                                    )
    assert np.abs(n - n_fit) < 3.7e-6
    assert np.abs(r - r_fit) < 1.3e-9
    assert np.abs(c[0] - c_fit[0]) < 2.7e-3
    assert np.abs(c[1] - c_fit[1]) < 2.7e-3
    assert np.allclose(qpi.pha, qpi_fit.pha, atol=0.0031, rtol=0)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



from os.path import dirname, realpath, exists
from distutils.command.build import build
from setuptools import setup, find_packages
import sys


author = u"Paul Müller"
authors = [author]
description = 'library for analyzing spherical objects in phase imaging'
name = 'qpsphere'
year = "2017"

sys.path.insert(0, realpath(dirname(__file__))+"/"+name)
from _version import version


class BuildIncludeBinaries(build):
    """Put binary files in resource directory for wheel builds"""
    def run(self):
        import qpsphere
        qpsphere.util.download_binaries(package_dir=True)
        # run the original build command
        super(BuildIncludeBinaries, self).run()


setup(
    name=name,
    author=author,
    author_email='dev@craban.de',
    url='https://github.com/RI-imaging/qpsphere',
    version=version,
    packages=find_packages(),
    package_dir={name: name},
    # Note: package data are not included in the source distribution
    # (manually excluded in MANIFEST.in)
    package_data={"qpsphere": ['resources/*']},
    cmdclass={"build": BuildIncludeBinaries},
    license="MIT",
    description=description,
    long_description=open('README.rst').read() if exists('README.rst') else '',
    install_requires=["appdirs",
                      "lmfit",
                      "matplotlib",
                      "nrefocus>=0.1.5",
                      "numpy>=1.12.0",
                      "scikit-image>=0.11.0",
                      "scipy>=0.18.0",
                      "qpimage",
                      ],
    setup_requires=['pytest-runner'],
    tests_require=["pytest"],
    python_requires='>=3.6, <4',
    keywords=["digital holographic microscopy",
              "optics",
              "quantitative phase imaging",
              "refractive index",
              "scattering",
              ],
    classifiers= [
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Intended Audience :: Science/Research'
                 ],
    platforms=['ALL'],
    )



import appdirs
import pathlib
from pkg_resources import resource_filename
import shutil

from .models import _bhfield

#: User's cache directory
CACHE_PATH = pathlib.Path(
    appdirs.AppDirs(appname="python-qpsphere").user_cache_dir)
#: Qpsphere package `resources` directory.
RESCR_PATH = pathlib.Path(resource_filename("qpsphere", "resources"))


def download_binaries(package_dir=False):
    """Download all binaries for the current platform

    Parameters
    ----------
    package_dir: bool
        If set to `True`, the binaries will be downloaded to the
        `resources` directory of the qpsphere package instead of
        to the users application data directory. Note that this
        might require administrative rights if qpsphere is installed
        in a system directory.

    Returns
    -------
    paths: list of pathlib.Path
        List of paths to binaries. This will always return binaries in
        the `resources` directory of the qpsphere package (if binaries
        are present there), in disregard of the parameter `package_dir`.
    """
    # bhfield
    # make sure the binary is available on the system
    paths = _bhfield.fetch.get_binaries()

    if package_dir:
        # Copy the binaries to the `resources` directory
        # of qpsphere.
        pdir = RESCR_PATH
        outpaths = []
        for pp in paths:
            target = pdir / pp.name
            if not target.exists():
                shutil.copy(pp, target)
            outpaths.append(target)
    else:
        outpaths = paths

    return outpaths


def remove_binaries(package_dir=False):
    """Remove all binaries for the current platform

    Parameters
    ----------
    package_dir: bool
        If True, remove all binaries from the `resources`
        directory of the qpsphere package. If False,
        remove all binaries from the user's cache directory.
    """
    paths = []

    if package_dir:
        pdir = RESCR_PATH
    else:
        pdir = CACHE_PATH

    for pp in pdir.iterdir():
        if pp.name != "shipped_resources_go_here":
            paths.append(pp)

    for pp in paths:
        pp.unlink()



import numpy as np
import scipy.interpolate as spintp
import time

from .. import models


class SpherePhaseInterpolator(object):
    def __init__(self, model, model_kwargs,
                 pha_offset=0, nrel=0.1, rrel=0.05, verbose=0):
        """Interpolation in-between modeled phase images

        Parameters
        ----------
        model: str
            Name of the light-scattering model
            (see :const:`qpsphere.models.available`)
        model_kwargs: dict
            Keyword arguments for the sphere model; must contain:

            - radius: float
                 Radius of the sphere [m]
            - sphere_index: float
                 Refractive index of the object
            - medium_index: float
                 Refractive index of the surrounding medium
            - wavelength: float
                 Vacuum wavelength of the imaging light [m]
            - pixel_size: float
                 Pixel size [m]
            - grid_size: tuple of floats
                 Resulting image size in x and y [px]
            - center: tuple of floats
                 Center position in image coordinates [px]
        pha_offset: float
            Phase offset added to the interpolation result
        nrel : float
            Determines the border of the interpolation range for the
            refractive index: [n-(n-nmed)*nrel, n+(n-nmed)*nrel]
            with n=model_kwargs["sphere_index"] and
            nmed=model_kwargs["medium_index"]
        rrel : float
            Determines the border of the interpolation range for the
            radius: [r*(1-rrel), r*(1+rrel)] with
            r=model_kwargs["radius"]
        verbose: int
            Increases verbosity.
        """
        self.verbose = verbose

        #: scattering model
        self.model = model
        #: scattering model function
        self.sphere_method = models.model_dict[model]
        #: scattering model keyword arguments
        self.model_kwargs = model_kwargs
        #: current sphere radius [m]
        self.radius = model_kwargs["radius"]
        #: current sphere index
        self.sphere_index = model_kwargs["sphere_index"]
        #: current background phase offset
        self.pha_offset = pha_offset
        #: current pixel offset in x
        self.posx_offset = model_kwargs["center"][0]
        #: current pixel offset in y
        self.posy_offset = model_kwargs["center"][1]
        #: half of current search interval size for refractive index
        self.dn = abs(
            (self.sphere_index - model_kwargs["medium_index"]) * nrel)
        #: half of current search interval size for radius [m]
        self.dr = self.radius * rrel

        # dictionary for determining if a new phase image
        # needs to be computed
        self._n_border = np.zeros((3, 3), dtype=float)
        self._r_border = np.zeros((3, 3), dtype=float)

        # border phase images
        self._border_pha = {}

    @property
    def params(self):
        """Current interpolation parameter dictionary"""
        par = {"radius": self.radius,
               "sphere_index": self.sphere_index,
               "pha_offset": self.pha_offset,
               "center": [self.posx_offset, self.posy_offset]
               }
        return par

    @property
    def range_n(self):
        """Current interpolation range of refractive index"""
        return self.sphere_index - self.dn, self.sphere_index + self.dn

    @property
    def range_r(self):
        """Current interpolation range of radius"""
        return self.radius - self.dr, self.radius + self.dr

    def compute_qpi(self):
        """Compute model data with current parameters

        Returns
        -------
        qpi: qpimage.QPImage
            Modeled phase data

        Notes
        -----
        The model image might deviate from the fitted image
        because of interpolation during the fitting process.
        """
        kwargs = self.model_kwargs.copy()
        kwargs["radius"] = self.radius
        kwargs["sphere_index"] = self.sphere_index
        kwargs["center"] = [self.posx_offset, self.posy_offset]
        qpi = self.sphere_method(**kwargs)
        # apply phase offset
        bg_data = np.ones(qpi.shape) * -self.pha_offset
        qpi.set_bg_data(bg_data=bg_data, which_data="phase")
        return qpi

    def get_border_phase(self, idn=0, idr=0):
        """Return one of nine border fields

        Parameters
        ----------
        idn: int
            Index for refractive index.
            One of -1 (left), 0 (center), 1 (right)
        idr: int
            Index for radius.
            One of -1 (left), 0 (center), 1 (right)
        """
        assert idn in [-1, 0, 1]
        assert idr in [-1, 0, 1]

        n = self.sphere_index + self.dn * idn
        r = self.radius + self.dr * idr

        # convert to array indices
        idn += 1
        idr += 1
        # find out whether we need to compute a new border field
        if self._n_border[idn, idr] == n and self._r_border[idn, idr] == r:
            if self.verbose > 3:
                print("Using cached border phase (n{}, r{})".format(idn, idr))
            # return previously computed field
            pha = self._border_pha[(idn, idr)]
        else:
            if self.verbose > 3:
                print("Computing border phase (n{}, r{})".format(idn, idr))
            kwargs = self.model_kwargs.copy()
            kwargs["radius"] = r
            kwargs["sphere_index"] = n
            kwargs["center"] = [self.posx_offset, self.posy_offset]
            tb = time.time()
            pha = self.sphere_method(**kwargs).pha
            if self.verbose > 2:
                print("Border phase computation time:",
                      self.sphere_method.__module__, time.time() - tb)
            self._border_pha[(idn, idr)] = pha
            self._n_border[idn, idr] = n
            self._r_border[idn, idr] = r

        return pha

    def get_phase(self, nintp=None, rintp=None,
                  delta_offset_x=0, delta_offset_y=0):
        """Interpolate from the border fields to new coordinates

        Parameters
        ----------
        nintp: float or None
            Refractive index of the sphere
        rintp: float or None
            Radius of sphere [m]
        delta_offset_x: float
            Offset in x-direction [px]
        delta_offset_y: float
            Offset in y-direction [px]

        Returns
        -------
        phase_intp: 2D real-valued np.ndarray
            Interpolated phase at the given parameters

        Notes
        -----
        Not all combinations are poosible, e.g.

        - One of nintp or rintp must be None
        - The current interpolation range must include the values
          for rintp and nintp
        """
        if nintp is None:
            nintp = self.sphere_index

        if rintp is None:
            rintp = self.radius

        assert (rintp == self.radius or nintp ==
                self.sphere_index), "Only r or n can be changed at a time."
        assert rintp >= self.radius - self.dr
        assert rintp <= self.radius + self.dr
        assert nintp >= self.sphere_index - \
            self.dn, "Out of range: {} !> {}".format(
                nintp, self.sphere_index - self.dn)
        assert nintp <= self.sphere_index + self.dn

        left = self.get_border_phase(0, 0)
        if rintp == self.radius:
            dist = nintp - self.sphere_index
            dmax = self.dn
            if dist < 0:
                righ = self.get_border_phase(-1, 0)
            else:
                righ = self.get_border_phase(1, 0)
        else:
            dist = rintp - self.radius
            dmax = self.dr
            if dist < 0:
                righ = self.get_border_phase(0, -1)
            else:
                righ = self.get_border_phase(0, 1)

        # make dist positive so that we are interpolating from left to right
        dist = np.abs(dist)
        # perform linear interpolation of data.
        phas = left + (righ - left) * dist / dmax

        # interpolation of lateral movement
        ti = time.time()
        ipphas = spintp.RectBivariateSpline(np.arange(phas.shape[0]),
                                            np.arange(phas.shape[1]),
                                            phas)

        if delta_offset_x != 0 or delta_offset_y != 0:
            # Shift the image. The offset values used here
            # are not self.posx_offset and self.posy_offset!
            # The offset values are added to the fields computed
            # with self.posx_offset and self.posy_offset.
            newx = np.arange(phas.shape[0]) + delta_offset_x
            newy = np.arange(phas.shape[1]) + delta_offset_y
            phas = ipphas(newx, newy)
        if self.verbose > 2:
            print("Interpolation time for {}: {}".format(
                  self.model, time.time() - ti))

        return phas + self.pha_offset
